#!/usr/bin/python3
#
# This is not a stable API; for use within autopkgtest only.
#
# Copyright 2024 Canonical Ltd.
# Copyright 2024 Simon McVittie
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import shutil
import sys
import subprocess
from abc import ABC, abstractmethod
from pathlib import Path
from collections.abc import Sequence


class Dependency(ABC):
    @abstractmethod
    def check(self) -> bool:
        raise NotImplementedError


class Executable(Dependency):
    def __init__(
        self,
        exe: str,
        package: str,
        *,
        fatal: bool = True,
        if_not_root: bool = False,
        if_not_systemd: bool = False,
    ) -> None:
        self.exe = exe
        self.fatal = fatal
        self.package = package
        self.if_not_root = if_not_root
        self.if_not_systemd = if_not_systemd

    def check(self) -> bool:
        qualifiers: list[str] = []

        if self.if_not_root:
            if os.geteuid() == 0:
                return True

            qualifiers.append(" (not required if run as root)")

        if self.if_not_systemd:
            if Path("/run/systemd/system").exists():
                return True

            qualifiers.append(" (not required if init is systemd)")

        if shutil.which(self.exe):
            return True

        if self.fatal:
            level = "ERROR"
        else:
            level = "WARNING"

        sys.stderr.write(
            "%s: command %s not found, provided by package %s%s\n"
            % (level, self.exe, self.package, "".join(qualifiers))
        )
        return not self.fatal


class FileDependency(Dependency):
    def __init__(
        self,
        path: Path,
        package: str,
        *,
        fatal: bool = True,
    ) -> None:
        self.fatal = fatal
        self.path = path
        self.package = package

    def check(self) -> bool:
        if self.path.exists():
            return True

        if self.fatal:
            level = "ERROR"
        else:
            level = "WARNING"

        sys.stderr.write(
            "%s: file %s not found, provided by package %s\n"
            % (level, self.path, self.package)
        )
        return not self.fatal


class DirectoryDependency(Dependency):
    def __init__(self, path: Path) -> None:
        self.path = path

    def check(self) -> bool:
        if not os.path.exists(self.path):
            sys.stderr.write(
                "INFO: directory %s not found, creating it now\n" % self.path
            )
            try:
                os.makedirs(self.path)
            except Exception as e:
                sys.stderr.write("ERROR: could not create directory: %s\n" % e)
                return False

        if not os.access(self.path, os.W_OK):
            sys.stderr.write("ERROR: no write access to %s\n" % self.path)
            return False

        if not os.path.isdir(self.path):
            sys.stderr.write("ERROR: %s is not a directory\n" % self.path)
            return False

        return True


class KvmDependency(Dependency):
    def __init__(
        self,
        *,
        if_exists: bool,
    ) -> None:
        self.if_exists = if_exists

    def check(self) -> bool:
        if os.access("/dev/kvm", os.W_OK):
            return True

        if not os.path.exists("/dev/kvm"):
            if self.if_exists:
                return True
            else:
                sys.stderr.write("ERROR: /dev/kvm does not exist\n")
                return False

        sys.stderr.write("ERROR: no permission to write /dev/kvm\n")
        return False


class SpaceRequirement(Dependency):
    """Check if disk at path has specified amount of free space"""

    def __init__(
        self,
        path: Path,
        space: int,
        *,
        fatal: bool = False,
    ) -> None:
        self.path = path
        self.space = space
        self.fatal = fatal

    def check(self) -> bool:
        if not os.path.exists(self.path):
            return False
        free_space = shutil.disk_usage(self.path).free

        if self.fatal:
            level = "ERROR"
        else:
            level = "WARNING"

        if free_space <= self.space:
            missing = self.space - free_space
            sys.stderr.write(
                "%s: Missing %s bytes of space in directory %s \n"
                % (level, missing, os.path.realpath(self.path))
            )
            return not self.fatal
        return True


class Package(Dependency):
    def __init__(self, name: str, fatal: bool = False):
        self.name = name
        self.fatal = fatal

    def check(self) -> bool:
        level = "ERROR" if self.fatal else "WARNING"
        dpkg_ret = subprocess.run(
            ["dpkg-query", "--showformat=${Status}", "--show", self.name],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            text=True,
        )
        if dpkg_ret.stdout != "install ok installed":
            sys.stderr.write(f"{level}: Missing package {self.name}\n")
            return not self.fatal

        return True


def check_dependencies(dependencies: Sequence[Dependency]):
    ok = True

    # Intentionally not short-circuiting: we want to list all missing
    # dependencies at once
    for dep in dependencies:
        if not dep.check():
            ok = False

    return ok



# autopkgtest_args is part of autopkgtest
# autopkgtest is a tool for testing Debian binary packages
#
# autopkgtest is Copyright (C) 2006-2016 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# See the file CREDITS for a full list of credits information (often
# installed as /usr/share/doc/autopkgtest/CREDITS).

import os
import re
import argparse

import adtlog
import testdesc
import adt_testbed

__all__ = ["parse_args"]

actions = None


def read_changes(parser, changes):
    """Expand a Debian .changes file into contained paths"""

    try:
        files = testdesc.parse_rfc822(changes).__next__()["Files"]
    except (StopIteration, KeyError):
        parser.error("%s is invalid and does not contain Files:" % changes)
    changes_dir = os.path.dirname(changes)
    paths = []
    for f in files.split():
        if f.endswith(".deb") or f.endswith(".dsc"):
            paths.append(os.path.join(changes_dir, f))
    return paths


def process_package_arguments(parser, args):
    """Check positional arguments and produce adt_run_args compatible actions list"""

    # TODO: This should be simplified now that the old adt_run_args CLI got dropped
    # Sort action list by deb << dsc, for a "do what
    # I mean" compatible adt_run_args action list

    global actions
    debsrc_action = None
    has_debs = False

    # expand .changes files
    packages = []
    for p in args.packages:
        if p.endswith(".changes"):
            packages += read_changes(parser, p)
        else:
            packages.append(p)

    def set_debsrc(p, kind, built_bin=None):
        nonlocal debsrc_action
        if debsrc_action:
            parser.error("You must specify only one source package to test")
        debsrc_action = (kind, p, built_bin)

    for p in packages:
        if p.endswith(".deb") and os.path.exists(p):
            actions.append(("binary", p, None))
            has_debs = True
        elif p.endswith(".dsc") and os.path.exists(p):
            set_debsrc(p, "source")
        elif re.match("[0-9a-z][0-9a-z.+-]+$", p):
            set_debsrc(p, "apt-source", False)
        elif os.path.isfile(os.path.join(p, "debian", "control")):
            if os.path.exists(os.path.join(p, "debian", "files")):
                set_debsrc(p, "built-tree", False)
            else:
                set_debsrc(p, "unbuilt-tree")
        elif os.path.isfile(os.path.join(p, "debian", "tests", "control")):
            # degenerate Debian source tree with only debian/tests
            set_debsrc(p, "built-tree", False)
        elif "://" in p:
            set_debsrc(p, "git-source")
        else:
            parser.error("%s is not a valid test package" % p)

    # if no source is given, check if the current directory is a source tree
    if not debsrc_action and os.path.isfile("debian/control"):
        if os.path.exists("debian/files"):
            set_debsrc(".", "built-tree", False)
        else:
            set_debsrc(".", "unbuilt-tree")

    if not debsrc_action:
        parser.error("You must specify source package to test")

    if has_debs:
        args.built_binaries = False

    if debsrc_action:
        # some actions above disable built binaries, for the rest use the CLI option
        if debsrc_action[2] is None:
            debsrc_action = (debsrc_action[0], debsrc_action[1], args.built_binaries)

        actions.append(debsrc_action)

    adtlog.debug("actions: %s" % actions)
    adtlog.debug("build binaries: %s" % args.built_binaries)


class ArgumentParser(argparse.ArgumentParser):
    """autopkgtest ArgumentParser

    It enables include files with '@' and trims whitespace from their lines.
    """

    def __init__(self, **kwargs):
        super().__init__(fromfile_prefix_chars="@", **kwargs)

    def convert_arg_line_to_args(self, arg_line):
        return [arg_line.strip()]


class AppendCommaSeparatedArg(argparse.Action):
    def __call__(self, parser, args, value, option_string=None):
        result = getattr(args, self.dest, [])
        result.extend(value.split(","))
        setattr(args, self.dest, result)


def parse_args(arglist=None):
    """Parse autopkgtest command line arguments.

    Return (options, actions, virt-server-args).
    """
    global actions
    actions = []

    usage = "%(prog)s [options] [testbinary ...] testsrc -- virt-server [options]"
    description = """Test installed binary packages using the tests in testsrc.

testsrc can be one of a:
 - Debian *.dsc source package
 - Debian *.changes file containing a .dsc source package (and possibly binaries to test)
 - Debian source package directory
 - apt source package name (through apt-get source)
 - Debian source package in git (url#branchname)

You can specify local *.deb packages to test."""

    epilog = """The -- argument separates the autopkgtest actions and options
from the virt-server which provides the testbed. See e. g. man autopkgtest-virt-qemu
for details."""

    parser = argparse.ArgumentParser(
        usage=usage,
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog,
        add_help=False,
    )

    # test specification
    g_test = parser.add_argument_group(
        "arguments for specifying and modifying the test"
    )
    g_test.add_argument(
        "--override-control",
        metavar="PATH",
        help="run tests from control file/manifest PATH instead of the source package",
    )
    g_test.add_argument(
        "--test-name",
        dest="only_tests",
        action="append",
        default=[],
        help="run only given test name (repeatable).",
    )
    g_test.add_argument(
        "--skip-test",
        dest="skip_tests",
        action="append",
        help="skip given test name (repeatable).",
    )
    g_test.add_argument(
        "-B",
        "--no-built-binaries",
        dest="built_binaries",
        action="store_false",
        default=True,
        help="do not build/use binaries from .dsc, git source, or unbuilt tree",
    )
    g_test.add_argument(
        "-a",
        "--test-architecture",
        metavar="ARCH",
        help="run tests for (and when asked, build binaries "
        "for) ARCH instead of the testbed's architecture.",
    )
    g_test.add_argument(
        "packages",
        nargs="*",
        help="testsrc source package and testbinary packages as above",
    )

    # logging
    g_log = parser.add_argument_group("logging options")
    g_log.add_argument(
        "-o",
        "--output-dir",
        help="Write test artifacts (stdout/err, log, debs, etc)"
        " to OUTPUT-DIR (must not exist or be empty)",
    )
    g_log.add_argument(
        "-l",
        "--log-file",
        dest="logfile",
        help="Write the log LOGFILE, emptying it beforehand,"
        " instead of using OUTPUT-DIR/log",
    )
    g_log.add_argument(
        "--summary-file",
        dest="summary",
        help="Write a summary report to SUMMARY, emptying it beforehand",
    )
    g_log.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        dest="verbosity",
        const=0,
        default=1,
        help="Suppress all messages from %(prog)s itself except for the test results",
    )

    # test bed setup
    g_setup = parser.add_argument_group("test bed setup options")
    g_setup.add_argument(
        "--setup-commands",
        metavar="COMMANDS_OR_PATH",
        action="append",
        default=[],
        help="Run these commands after opening the testbed; "
        "can be a string with the commands, or a file "
        "containing the commands",
    )
    g_setup.add_argument(
        "-U",
        "--apt-upgrade",
        dest="apt_upgrade",
        action="store_true",
        help="Run apt update/dist-upgrade/autoremove before the tests",
    )
    g_setup.add_argument(
        "--setup-commands-boot",
        metavar="COMMANDS_OR_PATH",
        action="append",
        default=[],
        help="Run these commands after --setup-commands, "
        "and also every time the testbed is rebooted",
    )
    g_setup.add_argument(
        "--add-apt-source",
        action="append",
        dest="add_apt_sources",
        metavar='"deb http://MIRROR SUITE COMPONENT..."',
        default=[],
        help="Enable additional apt sources",
    )
    g_setup.add_argument(
        "--add-apt-release",
        action="append",
        dest="add_apt_releases",
        metavar='"RELEASE"',
        default=[],
        help="Add apt source entries for RELEASE",
    )
    g_setup.add_argument(
        "--pin-packages",
        action="append",
        dest="pin_packages",
        metavar="RELEASE=pkgname,src:srcname,...",
        default=[],
        help="Enable testing with packages from a different release. "
        "Sets up apt pinning to use only those packages from RELEASE; "
        "src:srcname expands to all binaries of srcname.",
    )
    g_setup.add_argument(
        "--apt-pocket",
        action="append",
        metavar="POCKETNAME[=pkgname,src:srcname,...]",
        default=[],
        help="Enable additional apt source for release-POCKETNAME. "
        "If packages are given, set up apt pinning to use "
        "only those packages from release-POCKETNAME; src:srcname "
        " expands to all binaries of srcname",
    )
    g_setup.add_argument(
        "--apt-default-release",
        dest="apt_default_release",
        metavar="SUITENAME_OR_CODENAME",
        help="Define the default release for apt. For apt pinning to work "
        "properly, the default release must be set to the release "
        "that should provide the packages that are not pinned. For "
        "Debian and Ubuntu, this is normally automatically detected.",
    )
    g_setup.add_argument(
        "--no-apt-fallback",
        dest="enable_apt_fallback",
        action="store_false",
        default=True,
        help="Disable the apt fallback which is used with --apt-pocket or "
        "--pin-packages in case installation of dependencies fails due to "
        "strict pinning",
    )
    g_setup.add_argument(
        "--copy",
        metavar="HOSTFILE:TESTBEDFILE",
        action="append",
        default=[],
        help="Copy file or dir from host into testbed after opening",
    )
    g_setup.add_argument(
        "--env",
        metavar="VAR=value",
        action="append",
        default=[],
        help="Set arbitrary environment variable for builds and test",
    )
    g_setup.add_argument(
        "--ignore-restrictions",
        default=[],
        metavar="RESTRICTION[,RESTRICTION...]",
        action=AppendCommaSeparatedArg,
        help="Run tests even if these restrictions would normally prevent it",
    )
    g_setup.add_argument(
        "--insecure",
        action="store_true",
        default=False,
        help="Run without caring about security. "
        "Tests might corrupt the testbed, or escape from it. "
        "Other processes on the host or the testbed "
        "might be able to suborn the account running autopkgtest.",
    )

    # privileges
    g_priv = parser.add_argument_group("user/privilege handling options")
    g_priv.add_argument(
        "-u", "--user", help="run tests as USER (needs root on testbed)"
    )
    g_priv.add_argument(
        "--gain-root",
        dest="gainroot",
        help="Command to gain root during package build, "
        "passed to dpkg-buildpackage -r",
    )

    # debugging
    g_dbg = parser.add_argument_group("debugging options")
    g_dbg.add_argument(
        "-d",
        "--debug",
        action="store_const",
        dest="verbosity",
        const=2,
        help="Show lots of internal autopkgtest debug messages",
    )
    g_dbg.add_argument(
        "-s",
        "--shell-fail",
        action="store_true",
        help="Run a shell in the testbed after any failed build or test",
    )
    g_dbg.add_argument(
        "--shell",
        action="store_true",
        help="Run a shell in the testbed after every test",
    )

    # timeouts
    g_time = parser.add_argument_group("timeout options")
    g_time.add_argument("--timeout", type=int, default=0, help="Global timeout")
    for k, v in adt_testbed.timeouts.items():
        g_time.add_argument(
            "--timeout-" + k,
            type=int,
            dest="timeout_" + k,
            metavar="T",
            help="set %s timeout to T seconds (default: %us)" % (k, v),
        )
    g_time.add_argument(
        "--timeout-factor",
        type=float,
        metavar="FACTOR",
        default=1.0,
        help="multiply all default timeouts by FACTOR",
    )

    # locale
    g_loc = parser.add_argument_group("locale options")
    g_loc.add_argument(
        "--set-lang",
        metavar="LANGVAL",
        help="set LANG on testbed to LANGVAL (default: C.UTF-8",
    )

    # misc
    g_misc = parser.add_argument_group("other options")
    g_misc.add_argument(
        "--no-auto-control",
        dest="auto_control",
        action="store_false",
        default=True,
        help="Disable automatic test generation with autodep8",
    )
    g_misc.add_argument(
        "--build-parallel",
        metavar="N",
        help='Set "parallel=N" DEB_BUILD_OPTION for building '
        "packages (default: number of available processors)",
    )
    g_misc.add_argument(
        "--needs-internet",
        dest="needs_internet",
        choices=["run", "try", "skip"],
        default="run",
        help='Define how to handle the needs-internet restriction. With "try" '
        "tests with needs-internet restrictions will be run, but if they fail "
        'they will be treated as flaky tests. With "skip" these tests will be '
        'skipped immediately and will not be run. With "run" the restriction '
        "is basically ignored.",
    )
    g_misc.add_argument(
        "-V",
        "--validate",
        action="store_true",
        default=False,
        help="validate the test control file and exit",
    )
    g_misc.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="show this help message and exit",
    )

    # first, expand argument files
    file_parser = ArgumentParser(add_help=False)
    arglist = file_parser.parse_known_args(arglist)[1]

    # split off virt-server args
    try:
        sep = arglist.index("--")
    except ValueError:
        # still allow --help
        sep = None
        virt_args = None
    if sep is not None:
        virt_args = arglist[sep + 1 :]
        arglist = arglist[:sep]

    # parse autopkgtest options
    args = parser.parse_args(arglist)
    adtlog.verbosity = args.verbosity
    adtlog.debug("autopkgtest options: %s" % args)
    adtlog.debug("virt-runner arguments: %s" % virt_args)

    if not virt_args:
        parser.error("You must specify -- <virt-server>...")

    # autopkgtest-virt-* prefix can be skipped
    if (
        virt_args
        and "/" not in virt_args[0]
        and not virt_args[0].startswith("autopkgtest-virt-")
    ):
        virt_args[0] = "autopkgtest-virt-" + virt_args[0]

    process_package_arguments(parser, args)

    # verify --env validity
    for e in args.env:
        if "=" not in e:
            parser.error("--env must be KEY=value")

    if args.set_lang:
        args.env.append("LANG=" + args.set_lang)

    # set (possibly adjusted) timeout defaults
    adt_testbed.global_timeout = args.timeout
    for k in adt_testbed.timeouts:
        v = getattr(args, "timeout_" + k)
        if v is None:
            adt_testbed.timeouts[k] = int(adt_testbed.timeouts[k] * args.timeout_factor)
        else:
            adt_testbed.timeouts[k] = v

    # this timeout is for the virt server, so pass it down via environment
    os.environ["AUTOPKGTEST_VIRT_COPY_TIMEOUT"] = str(adt_testbed.timeouts["copy"])

    # if we have --setup-commands and it points to a file, read its contents
    for i, c in enumerate(args.setup_commands):
        c = adt_testbed.locate_setup_command(c)
        if os.path.exists(c):
            with open(c, encoding="UTF-8") as f:
                args.setup_commands[i] = f.read().strip()

    for i, c in enumerate(args.setup_commands_boot):
        c = adt_testbed.locate_setup_command(c)
        if os.path.exists(c):
            with open(c, encoding="UTF-8") as f:
                args.setup_commands_boot[i] = f.read().strip()

    # parse --copy arguments
    copy_pairs = []
    for arg in args.copy:
        try:
            (host, tb) = arg.split(":", 1)
        except ValueError:
            parser.error("--copy argument must be HOSTPATH:TESTBEDPATH: %s" % arg)
        if not os.path.exists(host):
            parser.error("--copy host path %s does not exist" % host)
        copy_pairs.append((host, tb))
    args.copy = copy_pairs

    # split multiline add_apt_sources into separate entries
    args.add_apt_sources = [
        s for line in args.add_apt_sources for s in line.splitlines()
    ]
    # strip, drop empty, drop comments, deduplicate
    args.add_apt_sources = [s.strip() for s in args.add_apt_sources]
    args.add_apt_sources = [s for s in args.add_apt_sources if s and s[0] != "#"]
    args.add_apt_sources = list(dict.fromkeys(args.add_apt_sources))

    return (args, actions, virt_args)



#!/usr/bin/python3
#
# This is not a stable API; for use within autopkgtest only.
#
# Part of autopkgtest.
# autopkgtest is a tool for testing Debian binary packages.
#
# Copyright 2006-2016 Canonical Ltd.
# Copyright 2016-2020 Simon McVittie
# Copyright 2017 Martin Pitt
# Copyright 2017 Jiri Palecek
# Copyright 2017-2018 Collabora Ltd.
# Copyright 2018 Thadeu Lima de Souza Cascardo
# Copyright 2019 Michael Biebl
# Copyright 2019 Raphaël Hertzog
#
# autopkgtest-virt-qemu was developed by
# Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# See the file CREDITS for a full list of credits information (often
# installed as /usr/share/doc/autopkgtest/CREDITS).

import errno
import fcntl
import json
import os
import re
import shutil
import socket
import subprocess
import sys
import tempfile
import time
from pathlib import Path
from typing import (
    Literal,
)
from collections.abc import Sequence

import VirtSubproc
import adtlog


class AllocatedPort(int):
    """This object represents a port number. However this port number also
    happens to be locked using POSIX lockf. If a process owning an
    AllocatedPort dies in a bad way, the lock is released, but the file
    is left behind. If the AllocatedPort is release()ed the file is gone as
    well.
    """

    _lockfd: int | None
    _lockfile: str | None

    def __new__(cls, value: int, lockfd: int, lockfile: str) -> "AllocatedPort":
        self = int.__new__(AllocatedPort, value)
        self._lockfd = lockfd
        self._lockfile = lockfile
        return self

    def release(self) -> None:
        if self._lockfile is not None:
            os.unlink(self._lockfile)
            self._lockfile = None
        if self._lockfd is not None:
            os.close(self._lockfd)
            self._lockfd = None

    __del__ = release


def find_free_port(start: int) -> AllocatedPort | Literal[0]:
    """Find an unused port in the range [start, start+50)
    Returns an AllocatedPort. The port reservation is valid until that
    object has been .release()d or garbage collected.
    """
    fd = None
    try:
        for p in range(start, start + 50):
            adtlog.debug("find_free_port: trying %i" % p)
            lockfile = "/tmp/autopkgtest-virt-qemu.port.%i" % p
            try:
                fd = os.open(lockfile, os.O_RDWR | os.O_CREAT)
                fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                # There is a little race condition. We may have opened
                # a previous file and managed to lock it, but
                # AllocatedPort.release may have gotten to delete it.
                if not os.access(lockfile, os.F_OK):
                    os.close(fd)
                    fd = None
                    adtlog.debug(f"find_free_port: lost race for {p}")
                    continue
                # The lock is now held until we close fd.
            except OSError:
                if fd is not None:
                    os.close(fd)
                    fd = None
                adtlog.debug("find_free_port: %i is locked" % p)
                continue

            try:
                s = socket.create_connection(("127.0.0.1", p))
                # if that works, the port is taken
                s.close()
                continue
            except OSError as e:
                if e.errno == errno.ECONNREFUSED:
                    adtlog.debug("find_free_port: %i is free" % p)
                    ret = AllocatedPort(p, fd, lockfile)
                    fd = None
                    return ret
                os.close(fd)
                fd = None
    finally:
        if fd is not None:
            os.close(fd)

    adtlog.debug("find_free_port: all ports are taken")
    return 0


def get_cpuflag() -> Sequence[str]:
    """Return QEMU cpu option list suitable for host CPU"""

    try:
        with open("/proc/cpuinfo") as f:
            for line in f:
                if line.startswith("flags"):
                    words = line.split()
                    if "vmx" in words:
                        adtlog.debug(
                            "Detected KVM capable Intel host CPU, enabling nested KVM"
                        )
                        return ["-cpu", "kvm64,+vmx,+lahf_lm"]
                    elif "svm" in words:  # AMD kvm
                        adtlog.debug(
                            "Detected KVM capable AMD host CPU, enabling nested KVM"
                        )
                        # FIXME: this should really be the one below
                        # for more reproducible testbeds, but nothing
                        # except -cpu host works
                        # return ['-cpu', 'kvm64,+svm,+lahf_lm']
                        return ["-cpu", "host"]
    except OSError as e:
        adtlog.warning("Cannot read /proc/cpuinfo to detect CPU flags: %s" % e)
        # fetching CPU flags isn't critical (only used to enable
        # nested KVM), so don't fail here

    return []


def get_host_time_zone() -> str:
    """Return the host system's time zone, or an empty string."""
    try:
        target = os.readlink("/etc/localtime")

        if "/zoneinfo/" in target:
            return target.split("/zoneinfo/")[-1]
    except OSError:
        pass

    try:
        with open("/etc/timezone") as f:
            for line in f:
                if line.startswith("#"):
                    continue
                line = line.strip()
                if line:
                    return line
    except OSError:
        pass

    return ""


class QemuImage:
    def __init__(
        self,
        file: str,
        format: str | None = None,
        readonly: bool = False,
    ) -> None:
        self.file = file
        self.overlay: str | None = None
        self.readonly = readonly

        if format is not None:
            self.format = format
        else:
            info = json.loads(
                VirtSubproc.check_exec(
                    [
                        "qemu-img",
                        "info",
                        "--output=json",
                        self.file,
                    ],
                    outp=True,
                    timeout=5,
                )
            )

            if "format" not in info:
                VirtSubproc.bomb("Unable to determine format of %s" % self.file)

            self.format = str(info["format"])

    def __str__(self) -> str:
        bits: list[str] = []

        if self.overlay is None:
            bits.append(f"file={self.file}")
            bits.append(f"format={self.format}")
        else:
            bits.append(f"file={self.overlay}")
            bits.append("format=qcow2")
            bits.append("cache=unsafe")

        bits.append("if=virtio")
        bits.append("discard=unmap")

        if self.readonly:
            bits.append("readonly=on")

        return ",".join(bits)


class QemuSession:
    """
    A running instance of qemu.
    """

    def __init__(
        self,
        images: Sequence[QemuImage | str],
        *,
        boot: str,
        cpus: int = 1,
        efi_code: Path | None = None,
        efi_data: Path | None = None,
        overlay: bool = False,
        overlay_dir: str | None = None,
        qemu_architecture: str,
        qemu_command: str,
        qemu_options: Sequence[str] = (),
        ram_size: int = 1024,
        workdir: str | None = None,
        user_shared_dirs: list[tuple[str, str, str]] | None = None,
    ) -> None:
        """
        Constructor.

        images: Disk images for the VM. The first image is assumed to be
            the bootable, writable root filesystem, and we actually boot a
            snapshot. The remaining images are assumed to be read-only.

        boot: bios, efi, ieee1275 or none
        cpus: Number of vCPUs
        efi_code: Path to e.g. /usr/share/OVMF/OVMF_CODE_4M.fd
        efi_data: Path to e.g. /usr/share/OVMF/OVMF_VARS_4M.fd
        overlay: If true, use a temporary overlay for first image
        overlay_dir: Store writable overlays here (default: workdir)
        qemu_architecture: Architecture, from qemu's vocabulary
        qemu_command: qemu executable
        qemu_options: Space-separated options for qemu
        ram_size: Amount of RAM in MiB
        workdir: Directory for temporary files (default: a random
            subdirectory of $TMPDIR)
        user_shared_dirs: List of tuples (hostdir, guestdir, tag) for host
            directories to mount in the guest using the given 9p tag
        """

        self.consoles = {"hvc0", "hvc1"}
        self.cpus = cpus
        self.images: list[QemuImage] = []
        self.overlay_dir = overlay_dir
        self.qemu_architecture = qemu_architecture
        self.qemu_command = qemu_command
        self.ram_size = ram_size
        self.ssh_port = find_free_port(10022)
        self.user_shared_dirs = user_shared_dirs or []

        if workdir is None:
            workdir = tempfile.mkdtemp(prefix="autopkgtest-qemu.")

        self.workdir: str | None = workdir
        os.chmod(workdir, 0o755)
        self.shareddir = os.path.join(workdir, "shared")
        os.mkdir(self.shareddir)

        for i, image in enumerate(images):
            if isinstance(image, QemuImage):
                self.images.append(image)
            else:
                assert isinstance(image, str)

                self.images.append(
                    QemuImage(
                        file=image,
                        format=None,
                        readonly=(i != 0),
                    )
                )

        if overlay:
            self.images[0].overlay = self.prepare_overlay(self.images[0])

        if self.ssh_port:
            adtlog.debug("Forwarding local port %i to VM ssh port 22" % self.ssh_port)
            nic_opt = ",hostfwd=tcp:127.0.0.1:%i-:22" % self.ssh_port
        else:
            nic_opt = ""

        # start QEMU
        argv = [
            self.qemu_command,
        ]

        if "-machine" not in qemu_options:
            # Some architectures do not have a default machine model
            if self.qemu_architecture == "aarch64":
                # gic-version=max is necessary to enable kvm support when
                # running natively on an arm64 host
                if os.uname()[4] == "aarch64":
                    argv.extend(["-machine", "type=virt,gic-version=max"])
                else:
                    argv.extend(["-machine", "virt"])
            elif self.qemu_architecture == "arm":
                argv.extend(["-machine", "virt"])
            elif self.qemu_architecture == "riscv64":
                argv.extend(["-machine", "virt,acpi=off"])
            elif self.qemu_architecture in ("i386", "x86_64") and boot == "efi":
                argv.extend(["-machine", "q35"])

        # Some architectures can only be run with certain CPUs
        if "-cpu" not in qemu_options:
            if self.qemu_architecture == "aarch64":
                if os.uname()[4] == "aarch64":
                    # Specifying a particular CPU doesn't seem to be
                    # possible when running natively with kvm
                    argv.extend(["-cpu", "host"])
                else:
                    # -cpu host is obviously not possible on a foreign
                    # CPU, -cpu max is unusably slow, but this works!
                    argv.extend(["-cpu", "cortex-a53"])
            # Ubuntu Questing+ needs RVA23 baseline profile, so at
            # least -cpu rva23s64.
            if self.qemu_architecture == "riscv64":
                argv.extend(["-cpu", "rva23s64"])

        argv.extend(
            [
                "-m",
                str(ram_size),
                "-smp",
                str(cpus),
                "-nographic",
                "-device",
                "virtio-net-pci,netdev=net0",
                "-netdev",
                "user,id=net0" + nic_opt,
                "-object",
                "rng-random,filename=/dev/urandom,id=rng0",
                "-device",
                "virtio-rng-pci,rng=rng0,id=rng-device0",
                "-monitor",
                "unix:%s,server=on,wait=off" % self.get_socket_path("monitor"),
                "-virtfs",
                (
                    "local,id=autopkgtest,path=%s,security_model=none,"
                    "mount_tag=autopkgtest"
                )
                % self.shareddir,
            ]
        )

        for hostdir, _, tag9p in self.user_shared_dirs:
            readonly = "off" if tag9p.startswith("rw") else "on"
            argv.extend(
                [
                    "-virtfs",
                    ("local,id=%s,path=%s,security_model=none,readonly=%s,mount_tag=%s")
                    % (tag9p, hostdir, readonly, tag9p),
                ]
            )

        for hvc in ("hvc0", "hvc1"):
            if hvc == "hvc0" and self.qemu_architecture == "ppc64le":
                # The first -serial argument shows up as /dev/hvc0 in
                # the VM on ppc64le, so identify it as such
                argv.extend(
                    [
                        "-serial",
                        "unix:%s,server=on,wait=off" % self.get_socket_path("hvc0"),
                    ]
                )
            else:
                if hvc == "hvc0" or self.qemu_architecture == "ppc64le":
                    argv.extend(
                        [
                            "-device",
                            "virtio-serial",
                        ]
                    )
                argv.extend(
                    [
                        "-chardev",
                        "socket,path=%s,server=on,wait=off,id=%s"
                        % (
                            self.get_socket_path(hvc),
                            hvc,
                        ),
                        "-device",
                        "virtconsole,chardev=%s" % hvc,
                    ]
                )

        if self.qemu_architecture in ("x86_64", "i386"):
            self.consoles.add("ttyS0")
            self.consoles.add("ttyS1")
            argv.extend(
                [
                    "-serial",
                    "unix:%s,server=on,wait=off" % self.get_socket_path("ttyS0"),
                    "-serial",
                    "unix:%s,server=on,wait=off" % self.get_socket_path("ttyS1"),
                ]
            )
        elif self.qemu_architecture == "ppc64le":
            # There are no emulated hardware serial ports in this qemu
            # implementation, only hypervisor virtual consoles
            pass
        else:
            # On ARM this ends up as ttyAMA0 rather than ttyS0, but that's
            # close enough.
            self.consoles.add("ttyS0")
            argv.extend(
                [
                    "-serial",
                    "unix:%s,server=on,wait=off" % self.get_socket_path("ttyS0"),
                ]
            )

        for i, image in enumerate(self.images):
            argv.append("-drive")
            argv.append("index=%d,%s" % (i, image))

        if efi_code is not None:
            argv.append("-drive")
            argv.append(f"if=pflash,format=raw,unit=0,read-only=on,file={efi_code}")

        if efi_data is not None:
            shutil.copy(str(efi_data), "%s/efivars.fd" % workdir)
            argv.append("-drive")
            argv.append("if=pflash,format=raw,unit=1,file=%s/efivars.fd" % workdir)

        if os.path.exists("/dev/kvm"):
            if QemuFactory.kvm_compatible(self.qemu_architecture):
                argv.append("-enable-kvm")

            # Enable nested KVM by default on x86_64
            if (
                os.uname()[4] == "x86_64"
                and self.qemu_architecture == "x86_64"
                and "-cpu" not in qemu_options
            ):
                argv += get_cpuflag()

        # pass through option to qemu
        if qemu_options:
            argv.extend(qemu_options)

        adtlog.debug("full qemu command-line: %s" % argv)

        self.subprocess: subprocess.Popen[bytes] | None = subprocess.Popen(
            argv,
            stdin=subprocess.DEVNULL,
            stdout=sys.stderr,
            stderr=subprocess.STDOUT,
        )

    def get_socket_path(self, name: str) -> str:
        assert self.workdir is not None
        return os.path.join(self.workdir, name)

    @property
    def monitor_socket(self) -> socket.socket:
        return VirtSubproc.get_unix_socket(self.get_socket_path("monitor"))

    def get_console_socket(self) -> socket.socket:
        for name in ("ttyS0", "hvc0"):
            if name in self.consoles:
                path = self.get_socket_path(name)
                return VirtSubproc.get_unix_socket(path)

        VirtSubproc.bomb("No console socket available")
        raise AssertionError  # not reached

    def prepare_overlay(
        self,
        image: QemuImage,
    ) -> str:
        """Generate a temporary overlay image"""

        # generate a temporary overlay
        if self.overlay_dir is not None:
            overlay = os.path.join(
                self.overlay_dir,
                os.path.basename(image.file) + ".overlay-%s" % time.time(),
            )
        else:
            workdir = self.workdir
            assert workdir is not None
            overlay = os.path.join(workdir, "overlay.img")

        adtlog.debug("Creating temporary overlay image in %s" % overlay)
        VirtSubproc.check_exec(
            [
                "qemu-img",
                "create",
                "-f",
                "qcow2",
                "-F",
                image.format,
                "-b",
                os.path.abspath(image.file),
                overlay,
            ],
            outp=True,
            timeout=300,
        )
        return overlay

    def cleanup(self, timeout_poweroff: int | None = None) -> int | None:
        ret = None

        if self.subprocess is not None:
            if timeout_poweroff is not None:
                try:
                    self.subprocess.wait(timeout_poweroff)
                except subprocess.TimeoutExpired:
                    pass
            self.subprocess.terminate()
            ret = self.subprocess.wait()
            self.subprocess = None

        if self.workdir is not None:
            shutil.rmtree(self.workdir)
            self.workdir = None

        return ret


class QemuFactory:
    """
    A factory object that decides which qemu command is required,
    and can start a qemu session.
    """

    def __init__(
        self,
        *,
        boot: str = "auto",
        dpkg_architecture: str | None = None,
        qemu_architecture: str | None = None,
        qemu_command: str | None = None,
        qemu_options: Sequence[str] = (),
    ) -> None:
        """
        Constructor.

        boot: auto, bios, efi, ieee1275 or none
        dpkg_architecture: Architecture, from dpkg's vocabulary
        qemu_architecture: Architecture, from qemu's vocabulary
        qemu_command: qemu executable
        qemu_options: Arbitrary options for qemu
        """

        if qemu_architecture is not None:
            self.qemu_architecture: str = qemu_architecture
        elif dpkg_architecture is not None:
            self.qemu_architecture = self.qemu_arch_for_dpkg_arch(dpkg_architecture)
        else:
            self.qemu_architecture = ""

        if not self.qemu_architecture and shutil.which("dpkg") is not None:
            dpkg_architecture = subprocess.check_output(
                ["dpkg", "--print-architecture"], text=True
            ).strip()
            self.qemu_architecture = self.qemu_arch_for_dpkg_arch(dpkg_architecture)

        if not self.qemu_architecture and qemu_command is not None:
            guess = self.qemu_arch_for_command(qemu_command)

            if guess is not None:
                self.qemu_architecture = guess

        if not self.qemu_architecture:
            self.qemu_architecture = self.qemu_arch_for_uname()

        self.arm_on_aarch64 = False

        if qemu_command is not None:
            self.qemu_command: str = qemu_command
        elif self.qemu_architecture == "arm" and os.uname()[4] == "aarch64":
            # aarch64 host systems can't kvm-accelerate qemu-system-arm,
            # so automatically switch to qemu-system-aarch64 in 32-bit mode
            # (emulating a 32-bit ARMv8a)
            self.qemu_architecture = "aarch64"
            self.qemu_command = "qemu-system-aarch64"
            self.arm_on_aarch64 = True
        else:
            self.qemu_command = self.qemu_command_for_arch(self.qemu_architecture)

        if boot == "auto":
            if self.qemu_architecture in ("i386", "x86_64"):
                # virt-qemu doesn't do anything special for BIOS boot,
                # but build-qemu does
                boot = "bios"
            elif self.qemu_architecture in ("arm", "aarch64", "riscv64"):
                # Handled specially below
                boot = "efi"
            elif self.qemu_architecture == "ppc64le":
                # virt-qemu doesn't do anything special for IEEE 1275
                # (OpenFirmware) boot, but build-qemu does
                boot = "ieee1275"
            else:
                adtlog.debug(
                    "autopkgtest does not know how to boot this machine, "
                    "hopefully nothing special needs to be done?"
                )
                boot = "none"

        self.boot = boot
        self.efi_code: Path | None = None
        self.efi_data: Path | None = None
        self.efi_package: str = ""

        if boot == "efi":
            if self.qemu_architecture == "x86_64":
                code = "/usr/share/OVMF/OVMF_CODE_4M.fd"
                data = "/usr/share/OVMF/OVMF_VARS_4M.fd"
                package = "ovmf"

                if not os.path.exists(code):
                    code = "/usr/share/OVMF/OVMF_CODE.fd"
                    data = "/usr/share/OVMF/OVMF_VARS.fd"
            elif self.qemu_architecture == "i386":
                code = "/usr/share/OVMF/OVMF32_CODE_4M.secboot.fd"
                data = "/usr/share/OVMF/OVMF32_VARS_4M.fd"
                package = "ovmf-ia32"
            elif (
                self.qemu_architecture == "arm"
                or self.arm_on_aarch64
                or (
                    self.qemu_architecture == "aarch64"
                    and "aarch64=off" in str(qemu_options)
                )
            ):
                code = "/usr/share/AAVMF/AAVMF32_CODE.fd"
                data = "/usr/share/AAVMF/AAVMF32_VARS.fd"
                package = "qemu-efi-arm"
            elif self.qemu_architecture == "aarch64":
                code = "/usr/share/AAVMF/AAVMF_CODE.fd"
                data = "/usr/share/AAVMF/AAVMF_VARS.fd"
                package = "qemu-efi-aarch64"
            elif self.qemu_architecture == "riscv64":
                code = "/usr/share/qemu-efi-riscv64/RISCV_VIRT_CODE.fd"
                data = "/usr/share/qemu-efi-riscv64/RISCV_VIRT_VARS.fd"
                package = "qemu-efi-riscv64"
            else:
                VirtSubproc.bomb("Unknown architecture for EFI boot")

            self.efi_code = Path(code)
            self.efi_data = Path(data)
            self.efi_package = package
        else:
            adtlog.debug(
                "Assuming nothing special needs to be done to set up "
                "firmware to boot this machine (boot method: %s)" % boot
            )

        self.qemu_options = list(qemu_options)

    def new_session(self, **kwargs) -> QemuSession:
        adtlog.debug("qemu architecture: %s" % self.qemu_architecture)
        adtlog.debug("qemu command: %s" % self.qemu_command)

        qemu_options = self.qemu_options

        if self.arm_on_aarch64 and "-cpu" not in qemu_options:
            qemu_options = ["-cpu", "host,aarch64=off"] + qemu_options

        return QemuSession(
            boot=self.boot,
            efi_code=self.efi_code,
            efi_data=self.efi_data,
            qemu_architecture=self.qemu_architecture,
            qemu_command=self.qemu_command,
            qemu_options=qemu_options,
            **kwargs,
        )

    @staticmethod
    def dpkg_arch_for_qemu_arch(qemu_architecture: str) -> str:
        return {
            "aarch64": "arm64",
            # alpha
            "arm": "armhf",  # or armel or arm
            "avr": "avr32",
            # cris not in dpkg-architecture -L
            # hppa
            # i386
            "loongarch64": "loong64",
            # m68k
            # microblaze not in dpkg-architecture -L
            # microblazeel not in dpkg-architecture -L
            # mips
            # mips64
            # mips64el
            # mipsel
            # moxie not in dpkg-architecture -L
            # nios2
            # or1k
            "ppc": "powerpc",
            # ppc64
            "ppc64le": "ppc64el",
            # riscv32 not in dpkg-architecture -L
            # riscv64
            # rx not in dpkg-architecture -L
            # s390x
            # sh4
            # sh4eb
            # sparc
            # sparc64
            # tricore not in dpkg-architecture -L
            "x86_64": "amd64",  # or x32
            "x86_64-microvm": "amd64",  # or x32
            # xtensa not in dpkg-architecture -L
            # xtensaeb not in dpkg-architecture -L
        }.get(qemu_architecture, qemu_architecture)

    @staticmethod
    def qemu_arch_for_command(qemu_command: str) -> str | None:
        known = [
            "aarch64",
            "alpha",
            "arm",
            "avr",
            "cris",
            "hppa",
            "i386",
            "loongarch64",
            "m68k",
            "microblaze",
            "microblazeel",
            "mips",
            "mips64",
            "mips64el",
            "mipsel",
            "moxie",
            "nios2",
            "or1k",
            "ppc",
            "ppc64",
            "ppc64le",
            "riscv32",
            "riscv64",
            "rx",
            "s390x",
            "sh4",
            "sh4eb",
            "sparc",
            "sparc64",
            "tricore",
            "x86_64",
            "x86_64-microvm",
            "xtensa",
            "xtensaeb",
        ]

        base = os.path.basename(qemu_command)

        for k in known:
            if base == "qemu-system-" + k:
                return k

        for k in known:
            if ("qemu-system-" + k) in base:
                return k

        return None

    # TODO: when we require qemu 8 (Debian 13, Ubuntu 24.04) we will be
    # able to rely on each dpkg architecture having a symlink
    # qemu-system-${dpkg_architecture} -> qemu-system-${qemu_architecture}
    # if different, allowing this to be simplified.
    @staticmethod
    def qemu_arch_for_dpkg_arch(dpkg_architecture: str) -> str:
        return {
            # alpha
            "amd64": "x86_64",
            # arm
            "arm64": "aarch64",
            "arm64ilp32": "aarch64",
            # armeb?
            "armel": "arm",
            "armhf": "arm",
            "avr32": "avr",
            # hppa
            # i386
            # ia64?
            "loong64": "loongarch64",
            # m32r?
            # m68k
            # mips
            # mips64
            # mips64el
            # mipsel
            # mipsn32?
            # mipsn32el?
            # mipsn32r6?
            # mipsn32r6el?
            # mipsr6?
            # mipsr6el?
            # nios2
            # or1k
            "powerpc": "ppc",
            "powerpcel": "ppc64le",  # ?
            # powerpcspe?
            # ppc64
            "ppc64el": "ppc64le",
            # riscv64
            "s390": "s390x",  # ?
            # s390x
            # sh3?
            # sh3eb?
            # sh4
            # sh4eb
            # sparc
            # sparc64
            # tilegx?
            "x32": "x86_64",
        }.get(dpkg_architecture, dpkg_architecture)

    @staticmethod
    def qemu_arch_for_uname(uname_m: str | None = None) -> str:
        uname_to_qemu_arch = {"i[3456]86$": "i386", "^arm": "arm"}

        if uname_m is None:
            uname_m = os.uname()[4]

        for pattern, arch in uname_to_qemu_arch.items():
            if re.match(pattern, uname_m):
                return arch
        else:
            # usually the qemu architecture is the same as Linux uname -m
            return uname_m

    @staticmethod
    def qemu_command_for_arch(qemu_architecture: str) -> str:
        return "qemu-system-" + qemu_architecture

    @staticmethod
    def kvm_compatible(qemu_architecture: str, uname_m: str | None = None) -> bool:
        if uname_m is None:
            uname_m = os.uname()[4]

        if re.match("^i[3456]86$", uname_m):
            uname_m = "i386"

        return qemu_architecture in {
            "x86_64": ["x86_64", "i386"],
            "i386": ["i386"],
            "aarch64": ["aarch64"],
            # According to https://wiki.alpinelinux.org/wiki/S390x/Installation
            "s390x": ["s390x"],
            "ppc64le": ["ppc64le"],
        }.get(uname_m, [])



"""Provide a fake package archive for testing"""

# (C) 2012-2015 Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import tempfile
import shutil
import os
import subprocess
import atexit
import textwrap


class Archive:
    def __init__(
        self,
        *,
        path=None,
        pooldir="pool",
        suite=None,
        codename=None,
        component=None,
        notautomatic=False,
        butautomaticupgrades=False,
        codename_symlink=False,
    ):
        """Construct a local package test archive.

        By default, Packages.gz is created in the root of the target directory,
        so that the apt source will just use '/' for the suite and no
        components. If suite, codename and component are given, Packages.gz
        will be placed into 'dists/suite/component/binary-<arch>/', and the
        Suite and Codename fields of the Release file of this archive will be
        set to the specified values.

        If the notautomatic argument is true, "NotAutomatic: yes" will be set
        in the Release file. In a similar fashion, the butautomaticupgrades
        argument controls ButAutomaticUpgrades.

        If codename_symlink is true, the archive will be accessible by using
        the specified codename in the APT sources. This is accomplished by
        creating a dists/codename -> dists/suite symlink. This is useful to
        create archives with a Debian-like layout, where "unstable" is also
        available via its codename "sid".

        The archive is initially empty. You can create new packages with
        create_deb(). self.path contains the path of the archive, and
        self.apt_source provides an apt source "deb" line.

        If path is None (default), it is kept in a temporary directory which
        gets removed when the Archive object gets deleted. Otherwise the given
        path is used, which is useful for creating multiple suites/components;
        then you should use a different pooldir.
        """
        if path:
            self.path = path
        else:
            self.path = tempfile.mkdtemp(prefix="testarchive.")
            atexit.register(shutil.rmtree, self.path)
        self.pooldir = pooldir

        arch = subprocess.check_output(
            ["dpkg", "--print-architecture"], text=True
        ).strip()
        self.suite = suite
        self.codename = codename
        self.component = component
        self.notautomatic = notautomatic
        self.butautomaticupgrades = butautomaticupgrades
        if suite or codename or component:
            assert suite and codename and component, (
                "must specify all of none of: suite, codename, component"
            )
            self.index_dir = os.path.join("dists", suite, component, "binary-" + arch)
            os.makedirs(os.path.join(self.path, self.index_dir))
            self.apt_source = "deb [trusted=yes arch={}] file://{} {} {}".format(
                arch,
                self.path,
                suite,
                component,
            )

            if codename_symlink and codename != suite:
                os.symlink(suite, os.path.join(self.path, "dists", codename))
        else:
            self.apt_source = "deb [trusted=yes arch={}] file://{} /".format(
                arch,
                self.path,
            )
            self.index_dir = ""

    def create_deb(
        self,
        name,
        version="1",
        architecture="all",
        dependencies={},
        description="test package",
        extra_tags={},
        files={},
        component="main",
        srcpkg=None,
        update_index=True,
    ):
        """Build a deb package and add it to the archive.

        The only mandatory argument is the package name. You can additionally
        specify the package version (default '1'), architecture (default
        'all'), a dictionary with dependencies (empty by default; for example
        {'Depends': 'foo, bar', 'Conflicts: baz'}, a short description
        (default: 'test package'), and arbitrary extra tags.

        By default the package is empty. It can get files by specifying a
        path -> contents dictionary in 'files'. Paths must be relative.
        Example: files={'etc/foo.conf': 'enable=true'}

        The newly created deb automatically gets added to the "Packages" index,
        unless update_index is False.

        Return the path to the newly created deb package, in case you only need
        the deb itself, not the archive.
        """
        d = tempfile.mkdtemp()
        os.mkdir(os.path.join(d, "DEBIAN"))
        with open(os.path.join(d, "DEBIAN", "control"), "w") as f:
            f.write(
                """Package: %s
Maintainer: Test User <test@example.com>
Version: %s
Priority: optional
Section: devel
Architecture: %s
"""
                % (name, version, architecture)
            )

            if srcpkg:
                f.write("Source: %s\n" % srcpkg)

            for k, v in dependencies.items():
                f.write(f"{k}: {v}\n")

            f.write(
                """Description: %s
 Test dummy package.
"""
                % description
            )

            for k, v in extra_tags.items():
                f.write(f"{k}: {v}\n")

        for path, contents in files.items():
            if isinstance(contents, bytes):
                mode = "wb"
            else:
                mode = "w"
            pathdir = os.path.join(d, os.path.dirname(path))
            if not os.path.isdir(pathdir):
                os.makedirs(pathdir)
            with open(os.path.join(d, path), mode) as f:
                f.write(contents)

        if srcpkg is None:
            srcpkg = name
        if srcpkg.startswith("lib"):
            prefix = srcpkg[:4]
        else:
            prefix = srcpkg[0]
        dir = os.path.join(self.path, self.pooldir, component, prefix, srcpkg)
        if not os.path.isdir(dir):
            os.makedirs(dir)

        debpath = os.path.join(dir, f"{name}_{version}_{architecture}.deb")
        subprocess.check_call(
            ["dpkg-deb", "-Zgzip", "-b", d, debpath],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        shutil.rmtree(d)
        assert os.path.exists(debpath)

        if update_index:
            self.update_index()

        return debpath

    def add_sources(self, name, binaries, version="1", component="main"):
        """Add source package entry to the Sources index"""

        if type(binaries) is not dict:
            binaries = {k: "any" for k in binaries}

        if name.startswith("lib"):
            prefix = name[:4]
        else:
            prefix = name[0]
        poolsubdir = os.path.join(self.pooldir, self.component, prefix, name)

        # create a dsc
        dir = os.path.join(self.path, self.pooldir, component, prefix, name)
        if not os.path.isdir(dir):
            os.makedirs(dir)
        dscpath = os.path.join(dir, f"{name}_{version}.dsc")
        with open(dscpath, mode="w"):
            # For now we only care about the file being present,
            # so let's create it empty.
            pass

        if self.index_dir:
            srcindex_dir = os.path.join(
                self.path, os.path.dirname(self.index_dir), "source"
            )
        else:
            srcindex_dir = self.index_dir
        if not os.path.isdir(srcindex_dir):
            os.makedirs(srcindex_dir)

        with open(os.path.join(srcindex_dir, "Sources"), "a") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    Package: {name}
                    Binary: {", ".join(binaries.keys())}
                    Version: {version}
                    Priority: optional
                    Architecture: any all
                    Format: 1.0
                    Directory: {poolsubdir}
                    Files:
                     d41d8cd98f00b204e9800998ecf8427e 0 {name}_{version}.dsc
                    Package-List:
                    """
                )
            )
            for b, a in binaries.items():
                f.write(f" {b} deb admin optional arch={a}\n")
            f.write("Standards-Version: 1.0\n\n")

    def update_index(self):
        """Update the Packages index and Release file.

        This usually gets done automatically by create_deb(), but needs to be
        done if you manually copy debs into the archive or call create_deb with
        update_index==False.
        """
        old_cwd = os.getcwd()
        try:
            os.chdir(self.path)
            with open(os.path.join(self.index_dir, "Packages"), "wb") as f:
                subprocess.check_call(
                    ["apt-ftparchive", "packages", self.pooldir], stdout=f
                )
            if self.suite:
                rp = os.path.join(self.path, "dists", self.suite, "Release")
                try:
                    os.unlink(rp)
                except OSError:
                    pass
                release_cmd = [
                    "apt-ftparchive",
                    "release",
                    "-o",
                    "APT::FTPArchive::Release::Suite=" + self.suite,
                    "-o",
                    "APT::FTPArchive::Release::Codename=" + self.codename,
                ]
                if self.notautomatic:
                    release_cmd += ["-o", "APT::FTPArchive::Release::NotAutomatic=yes"]
                if self.butautomaticupgrades:
                    release_cmd += [
                        "-o",
                        "APT::FTPArchive::Release::ButAutomaticUpgrades=yes",
                    ]
                release_cmd += [os.path.dirname(rp)]
                release = subprocess.check_output(release_cmd)
                with open(rp, "wb") as f:
                    f.write(release)
        finally:
            os.chdir(old_cwd)
        subprocess.check_call(["chmod", "-R", "a+rX", "--", self.path])


if __name__ == "__main__":
    r = Archive(suite="testy", component="main")
    r.create_deb("vanilla")
    r.create_deb("libvanilla0", srcpkg="vanilla")
    r.create_deb("chocolate", dependencies={"Depends": "vanilla"})
    print(r.apt_source)
    r.add_sources("vanilla", ["vanilla", "libvanilla0"])

    p = Archive(
        path=r.path,
        pooldir="pool-proposed",
        suite="testy-proposed",
        codename="testy",
        component="main",
    )
    p.create_deb("vanilla", "2")
    r.create_deb("libvanilla0", "2", srcpkg="vanilla")
    p.create_deb("chocolate", "2", dependencies={"Depends": "vanilla (>= 2)"})
    print(p.apt_source)
    p.add_sources("vanilla", ["vanilla", "libvanilla0"], "2")

    subprocess.call(["bash", "-i"], cwd=r.path)



# testdesc is part of autopkgtest
# autopkgtest is a tool for testing Debian binary packages
#
# autopkgtest is Copyright (C) 2006-2014 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# See the file CREDITS for a full list of credits information (often
# installed as /usr/share/doc/autopkgtest/CREDITS).

import string
import re
import errno
import os.path
import subprocess
import tempfile
from collections.abc import Iterable

import debian.deb822
import debian.debian_support
import debian.debfile

import adtlog

#
# Abstract test representation
#

known_restrictions = frozenset(
    {
        "allow-stderr",
        "breaks-testbed",
        "build-needed",
        "flaky",
        "isolation-container",
        "isolation-machine",
        "needs-internet",
        "needs-reboot",
        "needs-recommends",
        "needs-root",
        "needs-sudo",
        "rw-build-tree",
        "skip-foreign-architecture",
        "skip-not-installable",
        "skippable",
        "superficial",
    }
)


# Keys are restrictions
# Values are sets of either:
# - str: a capability that is required by the restriction
# - set of two or more str: at least one of these caps is required
#   by the restriction
RESTRICTIONS_REQUIRE_CAPS: dict[str, frozenset[str | frozenset[str]]] = {
    "breaks-testbed": frozenset({"revert-full-system"}),
    "needs-internet": frozenset({"has_internet"}),
    "needs-reboot": frozenset({"reboot"}),
    "needs-root": frozenset({"root-on-testbed"}),
    "isolation-container": frozenset(
        {
            frozenset({"isolation-container", "isolation-machine"}),
        }
    ),
    "isolation-machine": frozenset({"isolation-machine"}),
}


# Keys are restrictions
# Values are sets of dependency strings
RESTRICTIONS_IMPLY_DEPS = {
    "needs-sudo": frozenset({"passwd", "sudo"}),
}


PKGDATADIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))


class Unsupported(Exception):
    """Test cannot be run in the testbed"""

    def __init__(self, testname, message):
        self.testname = testname
        self.message = message

    def __str__(self):
        return f"Unsupported test {self.testname}: {self.message}"

    def report(self):
        adtlog.report(self.testname, "SKIP %s" % self.message)


class InvalidControl(Exception):
    """Test has invalid control data"""

    def __init__(self, testname, message):
        self.testname = testname
        self.message = message

    def __str__(self):
        return f"InvalidControl test {self.testname}: {self.message}"

    def report(self):
        adtlog.report(self.testname, "BROKEN %s" % self.message)


class Test:
    """Test description.

    This is only a representation of the metadata, it does not have any
    actions.
    """

    def __init__(
        self,
        name: str,
        path: str | None,
        command: str | None,
        restrictions: Iterable[str],
        features: Iterable[str],
        depends: list[list[str]],
        package_under_test_depends: list[list[str]],
    ) -> None:
        """Create new test description

        A test must have either "path" or "command", the respective other value
        must be None.

        @name: Test name
        @path: path to the test's executable, relative to source tree
        @command: shell command for the test code
        @restrictions, @features: sets of strings, as in README.package-tests
        @depends: string list of test dependencies (packages)
        @package_under_test_depends: string list with the subset of @depends
        that come from the source package we are testing
        """
        if "/" in name:
            raise Unsupported(name, "test name may not contain / character")

        if not ((path is None) ^ (command is None)):
            raise InvalidControl(name, "Test must have either path or command")

        self.name = name
        self.path = path
        self.command = command
        self.restrictions = set(restrictions)
        self.features = set(features)
        self.depends = list(depends)
        self.package_under_test_depends = package_under_test_depends

        for r in self.restrictions:
            for d in RESTRICTIONS_IMPLY_DEPS.get(r, frozenset()):
                if d not in self.depends:
                    self.depends.append([d])

        # None while test hasn't run yet; True: pass, False: fail
        self.result = None
        self.skipped = False
        adtlog.debug(
            'Test defined: name %s path %s command "%s" '
            "restrictions %s features %s depends %s "
            % (name, path, command, sorted(restrictions), sorted(features), depends)
        )

    def passed(self):
        """Mark test as passed"""

        self.result = True
        if "superficial" in self.restrictions:
            adtlog.report(self.name, "PASS (superficial)")
        else:
            adtlog.report(self.name, "PASS")

    def set_skipped(self, reason):
        """Mark test as skipped"""
        # This isn't called skipped() to avoid clashing with the boolean
        # attribute.

        self.skipped = True
        self.result = True
        adtlog.report(self.name, "SKIP " + reason)

    def failed(self, reason):
        """Mark test as failed"""

        self.result = False
        if "flaky" in self.restrictions:
            adtlog.report(self.name, "FLAKY " + reason)
        else:
            adtlog.report(self.name, "FAIL " + reason)

    def check_testbed_compat(
        self,
        caps: list[str],
        ignore_restrictions: list[str] | None = None,
    ) -> None:
        """Check for restrictions incompatible with test bed capabilities.

        Raise Unsupported exception if there are any.
        """
        effective = set(self.restrictions)
        if ignore_restrictions:
            effective -= set(ignore_restrictions)
        provided = set(caps)

        for r in effective:
            if r not in known_restrictions:
                raise Unsupported(self.name, "unknown restriction %s" % r)

        for r in effective:
            need_caps = RESTRICTIONS_REQUIRE_CAPS.get(r, frozenset())

            for c in need_caps:
                if isinstance(c, str):
                    if c not in caps:
                        raise Unsupported(
                            self.name,
                            ('Test restriction "%s" requires testbed capability "%s"')
                            % (r, c),
                        )
                else:
                    # must have at least one of the capabilities
                    assert isinstance(c, frozenset)
                    assert len(c) > 1

                    if not (c & provided):
                        cap_names = sorted('"%s"' % n for n in c)
                        needed = ", ".join(cap_names[:-1])
                        needed += " and/or " + cap_names[-1]

                        raise Unsupported(
                            self.name,
                            ('Test restriction "%s" requires testbed capability %s')
                            % (r, needed),
                        )


#
# Parsing for Debian source packages
#


def parse_rfc822(path):
    """Parse Debian-style RFC822 file

    Yield dictionaries with the keys/values.
    """
    try:
        f = open(path, encoding="UTF-8")
    except OSError as oe:
        if oe.errno != errno.ENOENT:
            raise
        return

    # filter out comments, python-debian doesn't do that
    # (http://bugs.debian.org/743174)
    lines = []
    for line in f:
        # completely ignore ^# as that breaks continuation lines
        if line.startswith("#"):
            continue
        # filter out comments which don't start on first column (Debian
        # #743174); entirely remove line if all that's left is whitespace, as
        # that again breaks continuation lines
        if "#" in line:
            line = line.split("#", 1)[0]
            if not line.strip():
                continue
        lines.append(line)
    f.close()

    for p in debian.deb822.Deb822.iter_paragraphs(lines):
        r = {}
        for field, value in p.items():
            # un-escape continuation lines
            v = "".join(value.split("\n")).replace("  ", " ")
            field = string.capwords(field)
            r[field] = v
        yield r


def _split_and_strip(string: str, delimiter: str = ",") -> list[str]:
    """Split the provided string on delimiter, strip the resulting
    elements and return a list of the non-empty ones."""

    return [x for x in (x.strip() for x in string.split(delimiter)) if x]


def _debian_check_unknown_fields(name, record):
    unknown_keys = set(record.keys()).difference(
        {
            "Tests",
            "Test-command",
            "Restrictions",
            "Features",
            "Depends",
            "Tests-directory",
            "Classes",
            "Architecture",
        }
    )
    if unknown_keys:
        raise Unsupported(name, "unknown field %s" % unknown_keys.pop())


def _architecture_is_concerned(
    architecture: str,
    architecture_restrictions: list[str],
) -> bool:
    """Wrapper around python-debian's architecture_is_concerned() with perl
    fallback in case python-debian is too old."""

    # TODO: Add Depends: python3-debian (>= 0.1.48) to bin:autopkgtest and drop this.

    if hasattr(debian.debian_support, "DpkgArchTable") and hasattr(
        debian.debian_support.DpkgArchTable, "architecture_is_concerned"
    ):
        dpkg_arch_table = debian.debian_support.DpkgArchTable.load_arch_table()
        return dpkg_arch_table.architecture_is_concerned(
            architecture,
            architecture_restrictions,
        )
    else:
        res = subprocess.run(
            [
                os.path.join(PKGDATADIR, "lib", "arch-is-concerned.pl"),
                architecture,
                " ".join(architecture_restrictions),
            ]
        )
        return res.returncode == 0


def _debian_packages_from_source(
    srcdir: str,
    test_arch: str,
    test_arch_is_foreign: bool,
) -> tuple[list[str], list[str]]:
    """Parse source tree in srcdir and return tuple with: list of all binary
    deb packages built from that source, and the same list but restricted to
    packages that build for test_arch."""

    packages: list[str] = []
    packages_for_test_arch: list[str] = []

    for st in parse_rfc822(os.path.join(srcdir, "debian/control")):
        if "Package" not in st:
            # source stanza
            continue
        # filter out udebs and similar stuff which aren't "real" debs
        if (
            st.get("Xc-package-type", "deb") != "deb"
            or st.get("Package-type", "deb") != "deb"
        ):
            continue
        binarydeb = st["Package"]
        arch = st["Architecture"]
        packages.append(binarydeb)
        if arch == "all":
            packages_for_test_arch.append(binarydeb)
        else:
            arch_list = arch.split()
            if _architecture_is_concerned(test_arch, arch_list):
                if test_arch_is_foreign:
                    binarydeb += ":" + test_arch
                packages_for_test_arch.append(binarydeb)

    return (packages, packages_for_test_arch)


def _filter_variables(deps: Iterable[str]) -> list[str]:
    deplist = []
    for dep in deps:
        if re.match(r"^\${", dep):
            continue
        dep = re.sub(r"\(.*\${\S*}.*\)", "", dep)
        deplist.append(dep.strip())
    return deplist


def _debian_recommends_from_source(srcdir: str) -> list[str]:
    deps = []
    for st in parse_rfc822(os.path.join(srcdir, "debian/control")):
        if "Recommends" in st:
            deps += _split_and_strip(st["Recommends"])

    # Because we parse the source d/control file instead of the binary
    # control file, we may run into variables. See bug 1008206 for
    # some discussion.
    deps = _filter_variables(deps)

    # The following check is a bit silly as there were apparently no
    # recommends, but still the test asked for it. Let's not fail on
    # that, but also let's not add the empty string.
    if deps == [""]:
        deps = []

    return deps


def _debian_build_deps_from_source(srcdir: str) -> list[list[str]]:
    deps = []
    for st in parse_rfc822(os.path.join(srcdir, "debian/control")):
        for key in ("Build-depends", "Build-depends-indep", "Build-depends-arch"):
            if key in st:
                altgr = _split_and_strip(st[key])
                deps += [_split_and_strip(g, "|") for g in altgr]

    # @builddeps@ should always imply build-essential
    deps.append(["build-essential"])
    return deps


dep_re = re.compile(
    r"(?P<package>[a-z0-9+-.]+)(?::[a-z0-9_-]+)?\s*"
    r"(\((?P<relation><<|<=|>=|=|>>)\s*(?P<version>[^\)]*)\))?"
    r"(\s*\[(?P<arch>[a-z0-9+-.! ]+)\])?$"
)


def _debian_check_dep(testname, dep):
    """Check a single Debian dependency"""

    dep = dep.strip()
    m = dep_re.match(dep)
    if not m:
        raise InvalidControl(
            testname, "Test Depends field contains an invalid dependency `%s'" % dep
        )
    if m.group("version"):
        try:
            debian.debian_support.NativeVersion(m.group("version"))
        except ValueError:
            raise InvalidControl(
                testname,
                "Test Depends field contains "
                "dependency `%s' with an "
                "invalid version" % dep,
            )
        except AttributeError:
            # too old python-debian, skip the check
            pass

    return (m.group("package"), m.group("version"))


def _expand_test_depends(
    testname: str,
    dep_str: str,
    srcdir: str,
    test_arch: str,
    test_arch_is_foreign: bool,
) -> tuple[list[list[str]], list[list[str]]]:
    """Split test dependencies (comma separated), validate their syntax, and
    expand @, @builddeps@ and @recommends@. Returns list of dependencies
    and the subset of those dependencies that come from the source
    package under test.

    If test_arch_is_foreign is True, then :test_arch architecture qualifiers
    are added to packages that are not Architecture: all.

    This may raise an InvalidControl exception if there are invalid
    dependencies.
    """
    deps = []
    package_under_test_deps = []
    (my_packages, my_packages_for_test_arch) = _debian_packages_from_source(
        srcdir,
        test_arch,
        test_arch_is_foreign,
    )

    for alt_group_str in _split_and_strip(dep_str):
        adtlog.debug("processing dependency %s" % alt_group_str)
        if alt_group_str == "@":
            for d in my_packages_for_test_arch:
                adtlog.debug("synthesised dependency %s from @" % d)
                deps.append([d])
                package_under_test_deps.append([d])
        elif alt_group_str == "@builddeps@":
            for daltgr in _debian_build_deps_from_source(srcdir):
                adtlog.debug(
                    "synthesised dependency %s from @builddeps@" % " | ".join(daltgr)
                )
                deps.append(daltgr)
        elif alt_group_str == "@recommends@":
            for d in _debian_recommends_from_source(srcdir):
                adtlog.debug("synthesised dependency %s from @recommends@" % d)
                deps.append([d])
        else:
            alt_group_list = []
            package_under_test_alternatives = []
            for dep in _split_and_strip(alt_group_str, "|"):
                (pkg, version) = _debian_check_dep(testname, dep)
                if (
                    test_arch_is_foreign
                    and f"{pkg}:{test_arch}" in my_packages_for_test_arch
                ):
                    dep = dep.replace(pkg, f"{pkg}:{test_arch}", 1)
                    pkg += f":{test_arch}"
                alt_group_list.append(dep)
                if pkg in my_packages:
                    package_under_test_alternatives.append(pkg)
            if len(alt_group_list) == len(package_under_test_alternatives):
                # We depend on x|y|z, and all of x, y and z are binary
                # packages built by the package under test
                adtlog.debug(
                    "marked alternatives %s as part of the package under test"
                    % package_under_test_alternatives
                )
                package_under_test_deps.append(package_under_test_alternatives)
            deps.append(alt_group_list)

    if test_arch_is_foreign and ["build-essential"] in deps:
        # Ensure we are installing the testbed-native version of
        # build-essential by specifying :native. This is needed because the
        # build-essential package is not Multi-Arch: foreign, so we need to
        # make sure we are always picking up the package for the testbed
        # architecture, not for the foreign test architecture.
        deps.remove(["build-essential"])
        deps.append(["build-essential:native"])
        # Install the right crossbuild-essential package.
        deps.append([f"crossbuild-essential-{test_arch}:native"])
        # Add install libc-dev and libstdc++-dev, as a workaround for #815172.
        deps.append([f"libc-dev:{test_arch}"])
        deps.append([f"libstdc++-dev:{test_arch}"])

    return (deps, package_under_test_deps)


def _autodep8(srcdir):
    """Generate control file with autodep8"""

    f = tempfile.NamedTemporaryFile(prefix="autodep8.")
    try:
        autodep8 = subprocess.Popen(
            ["autodep8"], cwd=srcdir, stdout=f, stderr=subprocess.PIPE
        )
    except OSError as e:
        adtlog.debug("autodep8 not available (%s)" % e)
        return None

    err = autodep8.communicate()[1].decode()
    if autodep8.returncode == 0:
        f.flush()
        f.seek(0)
        ctrl = f.read().decode()
        adtlog.debug("autodep8 generated control: -----\n%s\n-------" % ctrl)
        return f

    f.close()
    adtlog.debug(
        "autodep8 failed to generate control (exit status %i): %s"
        % (autodep8.returncode, err)
    )
    return None


def _matches_architecture(host_arch, arch_wildcard):
    try:
        subprocess.check_call(
            [
                "perl",
                "-mDpkg::Arch",
                "-e",
                "exit(!Dpkg::Arch::debarch_is(shift, shift))",
                host_arch,
                arch_wildcard,
            ]
        )
    except subprocess.CalledProcessError as e:
        # returns 1 if host_arch is not matching arch_wildcard; other
        # errors shouldn't be ignored
        if e.returncode != 1:
            raise
        return False
    return True


def _check_architecture(name, test_arch, architectures):
    """Check if test_arch is supported by the test architectures

    The architecture list comes in two variants, positive: only this
    arch is supported (arch may be a wildcard) and negative: this arch
    is not supported (arch may be a wildcard). If there is any
    positive arch, every arch not explicitly listed is skipped. Debian
    Policy 7.1 explains that for (Build-)Depends it's not allowed to
    mix positive and negative, so let's not do either. The list can
    also be empty. Empty and ["any"] are the same, "all" isn't
    allowed.
    """

    if "all" in architectures:
        raise Unsupported(name, "Arch 'all' not allowed in Architecture field")

    if len(architectures) == 0 or architectures == ["any"]:
        return

    any_negative = False
    any_positive = False
    for arch in architectures:
        if arch[0] == "!":
            any_negative = True
            if _matches_architecture(test_arch, arch[1:]):
                raise Unsupported(
                    name,
                    "Test declares architecture as not " + "supported: %s" % test_arch,
                )
        if arch[0] != "!":
            any_positive = True

    if any_positive:
        if any_negative:
            raise Unsupported(
                name,
                "It is not permitted for some archs to "
                + "be prepended by an exclamation mark while "
                + "others aren't",
            )
        arch_matched = False
        for arch in architectures:
            if _matches_architecture(test_arch, arch):
                arch_matched = True

        if not arch_matched:
            raise Unsupported(
                name,
                "Test lists explicitly supported "
                + "architectures, but the current architecture "
                + "%s isn't listed." % test_arch,
            )


def parse_debian_source(
    srcdir: str,
    testbed_caps: list[str],
    test_arch: str,
    test_arch_is_foreign: bool,
    control_path: str | None = None,
    auto_control: bool = True,
    ignore_restrictions: list[str] | None = None,
    only_tests: list[str] | None = None,
) -> tuple[list[Test], bool]:
    """Parse test descriptions from a Debian DEP-8 source dir

    @ignore_restrictions: If we would skip the test due to these restrictions,
                          run it anyway

    You can specify an alternative path for the control file (default:
    srcdir/debian/tests/control).

    Return (list of Test objects, some_skipped). If this encounters any invalid
    restrictions, fields, or test restrictions which cannot be met by the given
    testbed capabilities, the test will be skipped (and reported so), and not
    be included in the result.

    This may raise an InvalidControl exception.
    """
    some_skipped = False
    command_counter = 0
    tests = []
    if not ignore_restrictions:
        ignore_restrictions = []
    if not control_path:
        control_path = os.path.join(srcdir, "debian", "tests", "control")
        dtc_exists = os.path.exists(control_path)
        try_autodep8 = False

        if auto_control:
            if not dtc_exists:
                try_autodep8 = True
            else:
                dcontrol_path = os.path.join(srcdir, "debian", "control")
                for record in parse_rfc822(dcontrol_path):
                    testsuite = record.get("Testsuite", "")
                    if "autopkgtest-pkg-" in testsuite:
                        try_autodep8 = True
                    # We only want to look at the source section
                    break

        if try_autodep8:
            control = _autodep8(srcdir)
            if control is not None:
                control_path = control.name
            elif not dtc_exists:
                return ([], False)
        elif not dtc_exists:
            adtlog.debug("auto_control is disabled, and no regular tests")
            return ([], False)

    for record in parse_rfc822(control_path):
        command = None
        try:
            restrictions = set(record.get("Restrictions", "").replace(",", " ").split())

            # needs-recommends is deprecated, but until removed, let's
            # replace it with @recommends@
            if "needs-recommends" in restrictions:
                record["Depends"] = record.get("Depends", "@") + ",@recommends@"

            feature_test_name = None
            features = set()
            record_features = record.get("Features", "").replace(",", " ").split()
            for feature in record_features:
                details = feature.split("=", 1)
                if details[0] != "test-name":
                    features.add(feature)
                    continue
                if len(details) != 2:
                    # No value, i.e. a bare 'test-name'
                    raise InvalidControl("*", "test-name feature with no argument")
                if feature_test_name is not None:
                    raise InvalidControl("*", "only one test-name feature allowed")
                feature_test_name = details[1]
                features.add(feature)
            architectures = record.get("Architecture", "").replace(",", " ").split()

            if "Tests" in record:
                test_names = record["Tests"].replace(",", " ").split()
                if len(test_names) == 0:
                    raise InvalidControl("*", '"Tests" field is empty')
                (depends, package_under_test_depends) = _expand_test_depends(
                    test_names[0],
                    record.get("Depends", "@"),
                    srcdir,
                    test_arch,
                    test_arch_is_foreign,
                )
                if "Test-command" in record:
                    raise InvalidControl(
                        "*", 'Only one of "Tests" or "Test-Command" may be given'
                    )
                if feature_test_name is not None:
                    raise InvalidControl(
                        "*", "test-name feature incompatible with Tests"
                    )
                test_dir = record.get("Tests-directory", "debian/tests")

                for n in test_names:
                    try:
                        _debian_check_unknown_fields(n, record)
                        _check_architecture(n, test_arch, architectures)

                        test = Test(
                            n,
                            os.path.join(test_dir, n),
                            None,
                            restrictions,
                            features,
                            depends,
                            package_under_test_depends,
                        )
                        test.check_testbed_compat(testbed_caps, ignore_restrictions)
                    except Unsupported as u:
                        if (not only_tests) or n in only_tests:
                            u.report()
                            some_skipped = True
                    else:
                        tests.append(test)
            elif "Test-command" in record:
                command = record["Test-command"]
                (depends, package_under_test_depends) = _expand_test_depends(
                    command,
                    record.get("Depends", "@"),
                    srcdir,
                    test_arch,
                    test_arch_is_foreign,
                )
                if feature_test_name is None:
                    command_counter += 1
                    name = "command%i" % command_counter
                else:
                    name = feature_test_name
                _debian_check_unknown_fields(name, record)
                _check_architecture(name, test_arch, architectures)
                test = Test(
                    name,
                    None,
                    command,
                    restrictions,
                    features,
                    depends,
                    package_under_test_depends,
                )
                test.check_testbed_compat(testbed_caps, ignore_restrictions)
                tests.append(test)
            else:
                raise InvalidControl("*", 'missing "Tests" or "Test-Command" field')
        except Unsupported as u:
            if not only_tests:
                u.report()
                some_skipped = True

    return (tests, some_skipped)
